from . import tree as T

class BetaReductionVisitor(object):

    def visit_Schema_down(self, t: T.Schema) -> None:
        self.definitions = t.definitions

    def visit_Schema_up(self, t: T.Schema) -> T.Schema:
        return T.Schema(value=t.value, definitions=T.Definitions(values={}))

    def visit_Reference_down(self, t: T.Reference) -> T.Type:
        try:
            definition = self.definitions.values[t.value]
        except KeyError:
            raise ValueError(f"Missing definition for {t.value}")
        return definition.visit(self)


def reduce(t: T.Schema) -> T.Schema:
    return t.visit(BetaReductionVisitor())



from abc import ABC, abstractmethod
from typing import NamedTuple, Optional, Set, Tuple
import json
import jsonschema
import re
import logging

from .indent import indent

# Logger for Schema visitors
vlog = logging.getLogger("jsonschema_cn:visitor")


class Type(ABC):

    CONSTRUCTOR_KWARGS: Tuple[str, ...] = ()

    def __init__(self, **kwargs):
        self._jsonschema = None  # Will be filled as a cache on demand
        for name in self.CONSTRUCTOR_KWARGS:
            setattr(self, name, kwargs[name])

    @abstractmethod
    def to_jsonschema(self):
        pass

    @property
    def jsonschema(self):
        """Cached compilation of the corresponding jsonschema."""
        if self._jsonschema is None:
            self._jsonschema = self.to_jsonschema()
        return self._jsonschema

    def __str__(self):
        a = [k + "=" + str(getattr(self, k)) for k in self.CONSTRUCTOR_KWARGS]
        return f"{self.__class__.__name__}({', '.join(a)})"

    def prettily(self):
        """Return a unicode, pretty-printed representation of me."""

        def indent(text):
            return "\n".join(("    " + line) for line in text.splitlines())

        if not self.CONSTRUCTOR_KWARGS:
            return self.__class__.__name__
        else:
            acc = []
            for name in self.CONSTRUCTOR_KWARGS:
                value = getattr(self, name)
                # TODO go through lists and tuples
                r = value.prettily() if isinstance(value, Type) else repr(value)
                acc.append(f"{name} = {r}")
            return self.__class__.__name__ + "(\n" + indent("\n".join(acc)) + "\n)"

    __repr__ = __str__

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        for name in self.CONSTRUCTOR_KWARGS:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    def visit(self, visitor):
        x = self.visit_down(visitor)
        if x is self:
            return x.visit_up(visitor)
        else:
            return x

    def visit_down(self, visitor):
        vlog.debug("down %s", self)
        name = f"visit_{self.__class__.__name__}_down"
        method = getattr(visitor, name, None)
        if method is not None:
            vlog.debug("match down %s", name)
            r = method(self)
        else:
            r = None
        return r if r is not None else self

    def visit_up(self, visitor):
        vlog.debug("up %s", self)
        if visitor is None:
            return self
        else:
            names = [
                f"visit_{self.__class__.__name__}_up",
                f"visit_{self.__class__.__name__}",
                "visit",
            ]
            for name in names:
                method = getattr(visitor, name, None)
                if method is not None:
                    vlog.debug("match up %s", name)
                    return method(self)
            return self


class Schema(Type):

    CONSTRUCTOR_KWARGS = ("value", "definitions")

    def to_jsonschema(self, check_definitions=True, prune=True):
        r = self.value.jsonschema
        if isinstance(r, dict):  # Could also be `False`
            if self.definitions.values:
                r["definitions"] = self.definitions.jsonschema
                if prune:
                    self._prune(r)
            if check_definitions:
                self._check_definitions(r)
            # Recreate the dict to change keys order
            r2 = {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "$comment": str(self),
            }
            r2.update(r)
            r = r2
        return r

    def _check_definitions(self, schema):
        """Verify that all references have their definition."""
        occurring_references = self._get_dict_references(schema)
        for k in occurring_references:
            if k not in self.definitions.values.keys():
                raise ValueError(f"Missing definition for {k}")

    @classmethod
    def _prune(cls, schema: dict):
        """Remove unused definitions on a jsonschema (not a CN schema).
        May iterate more than once, because
        some references may be used in other definitions."""
        occurring_references = cls._get_dict_references(schema)
        eliminated_references = False
        while True:
            for k in list(schema["definitions"].keys()):
                # Freeze the keys into a list, so that `del` won't interfere
                if k not in occurring_references:
                    del schema["definitions"][k]
                    eliminated_references = True
            if eliminated_references:
                # Some unused references have been removed.
                # By removing their definitions, maybe some other
                # references became unused => try to prune again.
                occurring_references = cls._get_dict_references(schema)
                eliminated_references = False
            else:
                # Reached a fix-point, nothing left to eliminate
                return

    @classmethod
    def _get_dict_references(cls, x) -> Set[str]:
        """Extract every definition usage from a compiled jsonschema,
        so that it can be checked that they are all defined."""
        if isinstance(x, dict):
            if "$ref" in x:
                # Don't return immediately, there may be definitions in a "$ref".
                r = {x["$ref"].rsplit("/", 1)[-1]}
            else:
                r = set()
            for k, v in x.items():
                if k != "$ref":
                    r |= cls._get_dict_references(v)
            return r
        elif isinstance(x, list):
            return set.union(*(cls._get_dict_references(y) for y in x))
        else:
            return set()

    def _combine(self, other, op):
        args = []
        if isinstance(other, Schema):
            # Two schemas: combine main entries and merge defition dicts
            for schema in (self, other):
                content = schema.value
                if isinstance(content, Operator) and content.operator == op:
                    args.extend(content.values)  # Flatten associative operations
                else:
                    args.append(content)
            combined_content = Operator(operator=op, values=args)
            combined_defs = self.definitions | other.definitions
            return Schema(value=combined_content, definitions=combined_defs)
        elif op == "anyOf" and (
            isinstance(other, Definitions) or isinstance(other, dict)
        ):
            # Schema + additional definitions
            # TODO Maybe '+' is a more appropriate operator than '|'?
            if isinstance(other, dict):
                other = Definitions.from_dict(other)
            combined_defs = self.definitions | other
            return Schema(value=self.value, definitions=combined_defs)
        else:
            raise ValueError(
                "Schemas can only be combined with each other or with definitions"
            )

    def __and__(self, other):
        return self._combine(other, "allOf")

    def __or__(self, other):
        return self._combine(other, "anyOf")

    def validate(self, data=None):
        if data is None:  # Validate the schema itself
            jsonschema.Draft7Validator.check_schema(self.jsonschema)
        else:  # Validate a piece of data against the schema
            jsonschema.validate(
                data, self.jsonschema, format_checker=jsonschema.draft7_format_checker,
            )

    def __str__(self):
        if self.definitions.values:
            return indent(f"{self.value} {self.definitions}")
        else:
            return indent(self.value.__str__())

    def __repr__(self):
        return f"Schema('{self}')"

    def visit(self, visitor):
        s = super().visit_down(visitor)
        if s is not self:
            return s
        v = self.value.visit(visitor)
        d = self.definitions.visit(visitor)
        if v is self.value and d is self.definitions:
            s = self
        else:
            s = self.__class__(value=v, definitions=d)
        return s.visit_up(visitor)


class Definitions(Type):

    CONSTRUCTOR_KWARGS = ("values",)

    def to_jsonschema(self):
        return {k: v.jsonschema for k, v in self.values.items()}

    def __or__(self, other):
        if isinstance(other, Definitions) or isinstance(other, dict):
            if isinstance(other, dict):
                other = Definitions.from_dict(other)
            overlap = set(self.values.keys()) & set(other.values.keys())
            conflicts = sorted(
                name for name in overlap if self.values[name] != other.values[name]
            )
            if conflicts:
                raise ValueError(
                    f"Cannot merge definitions, conflict over {', '.join(conflicts)}"
                )
            defs = dict(self.values)
            defs.update(other.values)
            return Definitions(values=defs)
        elif isinstance(other, Schema):
            return other | self
        else:
            raise ValueError("Cannot perform 'or' on Definitions and that")

    @staticmethod
    def from_dict(d):
        definitions = Definitions(values={})
        for name, schema in d.items():
            if isinstance(schema, str):  # Convert src strings -> Schema
                from .peg_visitor import parse

                schema = parse("schema", schema)
            definitions |= schema.definitions
            definitions |= Definitions(values={name: schema.value})
        return definitions

    def __str__(self):
        if self.values:
            return "\nwhere " + "\nand ".join(f"{k} = {v}" for k, v in self.values.items())
        else:
            return "<empty definitions>"

    def visit(self, visitor):
        s = super().visit_down(visitor)
        if s is not self:
            return s
        visited = {name: type.visit(visitor) for name, type in self.values.items()}
        if any(a is not b for a, b in zip(visited.values(), self.values.values())):
            s = self.__class__(values=visited)
        else:
            s = self
        return s.visit_up(visitor)


class Integer(Type):

    CONSTRUCTOR_KWARGS = ("cardinal", "multiple")

    def to_jsonschema(self):
        (card_min, card_max) = self.cardinal
        r = {"type": "integer"}
        if card_min is not None:
            r["minimum"] = card_min
        if card_max is not None:
            r["maximum"] = card_max
        if self.multiple is not None:
            r["multipleOf"] = self.multiple
        return r

    def __str__(self):
        r = "integer"
        (card_min, card_max) = self.cardinal
        if card_min is not None and card_max is not None:
            r += "{" + str(card_min) + ", " + str(card_max) + "}"
        elif card_min is not None:
            r += "{" + str(card_min) + ", _}"
        elif card_max is not None:
            r += "{_, " + str(card_max) + "}"
        if self.multiple is not None:
            r += f"/{self.multiple}"
        return r


class String(Type):

    CONSTRUCTOR_KWARGS = ("cardinal", "format", "regex")

    def to_jsonschema(self):
        r = {"type": "string"}
        card_min, card_max = self.cardinal
        if card_min is not None:
            r["minLength"] = card_min
        if card_max is not None:
            r["maxLength"] = card_max
        if self.format is not None:
            r["format"] = self.format
        if self.regex is not None:
            r["pattern"] = self.regex
        return r

    def __str__(self):
        if self.format is not None:
            r = f'f"{self.format}"'
        elif self.regex is not None:
            r = f'r"{self.regex}"'  # TODO What about escaping special chars?
        else:
            r = "string"
        (card_min, card_max) = self.cardinal
        if card_min is not None and card_max is not None:
            if card_min == card_max:
                r += "{" + str(card_min) + "}"
            else:
                r += "{" + str(card_min) + ", " + str(card_max) + "}"
        elif card_min is not None:
            r += "{" + str(card_min) + ", _}"
        elif card_max is not None:
            r += "{_, " + str(card_max) + "}"
        return r


class Forbidden(Type):
    CONSTRUCTOR_KWARGS = ()

    def to_jsonschema(self):
        return False

    def __str__(self):
        return "forbidden"


class Litteral(Type):
    CONSTRUCTOR_KWARGS = ("value",)

    def to_jsonschema(self):
        return {"type": self.value}

    def __str__(self):
        return self.value


class Constant(Type):
    CONSTRUCTOR_KWARGS = ("value",)

    def to_jsonschema(self):
        return {"const": self.value}

    def __str__(self):
        if isinstance(self.value, str):
            # with double-quotes and escapes but no backticks
            return json.dumps(self.value)
        else:
            return f"`{json.dumps(self.value)}`"


class Operator(Type):
    CONSTRUCTOR_KWARGS = ("operator", "values")

    def to_jsonschema(self):
        return {self.operator: [v.jsonschema for v in self.values]}

    def __str__(self):
        # TODO Different operator for oneOf/anyOf
        op = {"anyOf": " | ", "oneOf": " | ", "allOf": " & "}[self.operator]
        return op.join(v.__str__() for v in self.values)

    def visit(self, visitor):
        s = self.visit_down(visitor)
        if s is not self:
            return s
        visited = [c.visit(visitor) for c in self.values]
        if any(a is not b for a, b in zip(visited, self.values)):
            s = self.__class__(operator=self.operator, values=visited)
        else:
            s = self
        return s.visit_up(visitor)


class Not(Type):
    CONSTRUCTOR_KWARGS = ("value",)

    def to_jsonschema(self):
        return {"not": self.value.jsonschema}

    def __str__(self):
        return f"not {self.value}"

    def visit(self, visitor):
        s = self.visit_down(visitor)
        if s is not self:
            return s
        v = self.value.visit(visitor)
        if v is self.value:
            s = self
        else:
            s = self.__class__(value=v)
        return s.visit_up(visitor)


class Enum(Type):
    CONSTRUCTOR_KWARGS = ("values",)

    def to_jsonschema(self):
        return {"enum": list(self.values)}

    def __str__(self):
        def c2s(v):
            return json.dumps(v) if isinstance(v, str) else f"`{json.dumps(v)}`"
        return " | ".join(c2s(v) for v in self.values)


class Reference(Type):
    CONSTRUCTOR_KWARGS = ("value",)

    def to_jsonschema(self):
        return {"$ref": "#/definitions/" + self.value}

    def __str__(self):
        return f"<{self.value}>"


class ObjectProperty(NamedTuple):
    name: Optional[str]
    optional: bool
    type: Type


class Object(Type):

    CONSTRUCTOR_KWARGS = (
        "properties",
        "cardinal",
        "additional_property_types",
        "additional_property_names",
    )

    def to_jsonschema(self):
        card_min, card_max = self.cardinal
        r = {"type": "object"}
        properties = {}
        required = []
        for (k, opt, v) in self.properties:
            if not opt:
                required.append(k)
            properties[k] = v.jsonschema if v is not None else True
        if required:
            r["required"] = required
        if properties:
            r["properties"] = properties

        if self.additional_property_types is False:
            r["additionalProperties"] = False
        elif self.additional_property_types is not None:
            r["additionalProperties"] = self.additional_property_types.jsonschema
        if self.additional_property_names is not None:
            r["propertyNames"] = self.additional_property_names.jsonschema

        implicit_card_min = len(required)
        implicit_card_max = (
            len(properties) if r.get("additionalProperties") is False else None
        )
        if card_min is not None:
            if card_min > implicit_card_min:
                r["minProperties"] = card_min
            if implicit_card_max is not None and card_min > implicit_card_max:
                raise ValueError(
                    f"Can only have up to {implicit_card_max} properties, not {card_min}"
                )
        if card_max is not None:
            if implicit_card_max is None or card_max < implicit_card_max:
                r["maxProperties"] = card_max
            if implicit_card_min > card_max:
                raise ValueError(
                    f"Must have at least {implicit_card_min} properties, which is more than {card_max}"
                )
        return r

    def __str__(self):
        if self.additional_property_names and self.additional_property_types:
            only = f"only {self.additional_property_names}: {self.additional_property_types}"
        elif self.additional_property_names:
            only = f"only {self.additional_property_names}"
        elif self.additional_property_types is True:
            only = f"only _: {self.additional_property_types}"
        elif self.additional_property_types is False:
            only = f"only"
        else:
            only = None
        if self.properties:

            def needs_quotes(name):
                return name in ("only", "unique") or not re.match(r"^\w+$", name)

            def pair_str(item):
                (name, opt, t) = item
                opt = "?" if opt else ""
                if needs_quotes(name):
                    name = json.dumps(name)
                t = "_" if t is None else t.__str__()
                return f"{name}{opt}: {t}"

            properties = ",\n".join(pair_str(item) for item in self.properties)
        else:
            properties = None
    
        multi_lines = len(self.properties) > 1
        if only == "only":
            r = "only " + properties if properties else "only"
        elif only is None:
            r = properties or ""
        elif properties is not None:
            r = only + ",\n" + properties
            multi_lines = True
        else:
            r = only
        
        r = "{\n" + r + "\n}" if multi_lines else "{" + r + "}"
        (card_min, card_max) = self.cardinal
        if card_min is not None and card_max is not None:
            if card_min == card_max:
                r += "{" + str(card_min) + "}"
            else:
                r += "{" + str(card_min) + ", " + str(card_max) + "}"
        elif card_min is not None:
            r += "{" + str(card_min) + ", _}"
        elif card_max is not None:
            r += "{_, " + str(card_max) + "}"
        return r

    def visit(self, visitor):
        s = self.visit_down(visitor)
        if s is not self:
            return s
        visited_props = [
            ObjectProperty(p[0], p[1], p[2].visit(visitor)) for p in self.properties
        ]

        apn = self.additional_property_names
        if isinstance(apn, Type):
            apn = apn.visit(visitor)

        apt = self.additional_property_types
        if isinstance(apt, Type):
            apt = apt.visit(visitor)

        if apt is not self.additional_property_types or \
           apn is not self.additional_property_names or \
           any(a[2] is not b[2] for a, b in zip(visited_props, self.properties)):
            s = self.__class__(
                properties=visited_props,
                additional_property_names=apn,
                additional_property_types=apt,
                cardinal=self.cardinal,
            )
        else:
            s = self
        return s.visit_up(visitor)


class Array(Type):

    CONSTRUCTOR_KWARGS = ("items", "additional_items", "cardinal", "unique")

    def to_jsonschema(self):
        # types = self.kwargs["types"]
        # extra_type = self.kwargs["additional_types"]
        # card_min, card_max = self.kwargs["cardinal"]
        r = {"type": "array"}

        if self.items:  # Tuple array
            r["items"] = [item.jsonschema for item in self.items]
            if self.additional_items is False:  # No extra items allowed
                r["additionalItems"] = False
            elif self.additional_items is True:  # Extra items with any type
                pass
            else:  # extra items allowed, but wiht a constrained type
                r["additionalItems"] = self.additional_items.jsonschema
        elif isinstance(
            self.additional_items, Type
        ):  # List array with homogeneous type
            r["items"] = self.additional_items.jsonschema

        card_min, card_max = self.cardinal
        implicit_card_min = len(self.items)
        implicit_card_max = len(self.items) if self.additional_items is False else None
        if card_min is not None and card_min > implicit_card_min:
            r["minItems"] = card_min
            if implicit_card_max is not None and card_min > implicit_card_max:
                raise ValueError(
                    f"Can only have up to {implicit_card_max} items, not {card_min}"
                )
        if card_max is not None:
            if implicit_card_min > card_max:
                raise ValueError(
                    f"Must have at least {implicit_card_min} items, which is more than {card_max}"
                )
            if implicit_card_max is None or card_max < implicit_card_max:
                r["maxItems"] = card_max

        if self.unique:
            r["uniqueItems"] = True

        return r

    def __str__(self):
        if self.additional_items is False:
            prefix = f"only "
        else:
            prefix = ""
        if self.unique:
            prefix += "unique "
        items = [str(item) for item in self.items]
        if isinstance(self.additional_items, Type):
            items.append(f"{self.additional_items}*")
        items = ", ".join(items)
        r = f"[{prefix}{items}]"
        if prefix and items:
            r = f"[{prefix}{items}]"
        elif prefix:
            r = f"[{prefix.strip()}]"
        elif items:
            r = f"[{items}]"
        else:
            r = "[ ]"
        (card_min, card_max) = self.cardinal
        if card_min is not None and card_max is not None:
            if card_min == card_max:
                r += "{" + str(card_min) + "}"
            else:
                r += "{" + str(card_min) + ", " + str(card_max) + "}"
        elif card_min is not None:
            r += "{" + str(card_min) + ", _}"
        elif card_max is not None:
            r += "{_, " + str(card_max) + "}"
        return r

    def visit(self, visitor):
        s = self.visit_down(visitor)
        if s is not self:
            return s
        visited_items = [c.visit(visitor) for c in self.items]
        if isinstance(self.additional_items, Type):
            additems = self.additional_items.visit(visitor)
        else:
            additems = self.additional_items
        if additems is not self.additional_items or any(
            a is not b for a, b in zip(visited_items, self.items)
        ):
            s = self.__class__(
                items=visited_items,
                additional_items=additems,
                unique=self.unique,
                cardinal=self.cardinal,
            )
        else:
            s = self
        return s.visit_up(visitor)


class Conditional(Type):

    CONSTRUCTOR_KWARGS = ("if_term", "then_term", "else_term")

    def to_jsonschema(self):
        r = {"if": self.if_term.jsonschema, "then": self.then_term.jsonschema}
        if self.else_term:
            r["else"] = self.else_term.jsonschema
        return r

    def __str__(self):
        r = f"if {self.if_term} then {self.then_term}"
        if self.else_term is not None:
            r += f" {self.else_term}"
        return r

    def visit(self, visitor):
        s = self.visit_down(visitor)
        if s is not self:
            return s
        if_t = self.if_term.visit(visitor)
        then_t = self.then_term.visit(visitor)
        if self.else_term is not None:
            else_t = self.else_term.visit(visitor)
        else:
            else_t = None
        if (
            if_t is not self.if_term
            or then_t is not self.then_term
            or else_t is not self.else_term
        ):
            s = self.__class__(if_term=if_t, then_term=then_t, else_term=else_t)
        else:
            s = self
        return s.visit_up(visitor)



OPENING = set("[{(")
CLOSING = set("]})")

def indent(src, step=2):
    i = 0
    output = []
    for line in src.split("\n"):
        line = line.strip()
        j = 0
        for k in line:
            if k.isspace():
                pass
            elif k in CLOSING:
                i -= 1
            else:
                break
            j += 1
        output.append(" " * (i*step) + line)
        for k in line[j:]:
            if k in OPENING:
                i += 1
            elif k in CLOSING:
                i -= 1
    return "\n".join(output)



import sys
import json
from argparse import ArgumentParser

from .peg_visitor import parse


def main():
    parser = ArgumentParser(
        description="Convert from a compact DSL into full JSON schema."
    )
    parser.add_argument(
        "-o", "--output", default="-", help="Output file; defaults to stdout"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=True,
        default=False,
        help="Verbose output",
    )
    parser.add_argument(
        "-r",
        "--reduce",
        action="store_const",
        const=True,
        default=False,
        help="Reduce variable references",
    )
    parser.add_argument(
        "-f",
        "--format",
        action="store_const",
        const=True,
        default=False,
        help="Reformat instead of compiling",
    )
    parser.add_argument(
        "filename",
        nargs="?",
        default="-",
        help="Input file; use '-' to read from stdin.",
    )
    parser.add_argument(
        "--version",
        action="store_const",
        const=True,
        default=False,
        help="Display version and exit",
    )
    args = parser.parse_args()

    if args.version:
        from . import __version__
        print(f"JSON Schema compact notation v{__version__}.")
        exit(0)

    input = open(args.filename, "r") if args.filename != "-" else sys.stdin
    output = open(args.output, "w") if args.output != "-" else sys.stdout

    source = input.read()
    schema = parse("schema", source, verbose=args.verbose)

    if args.reduce:
        from .beta import reduce as reduce
        schema = reduce(schema)

    if args.format:
        result = str(schema)
    else:
        try:
            # TODO Try and guess TTY width
            import jsview
            result = jsview.dumps(schema.jsonschema)
        except ModuleNotFoundError:
            # Raw printing if jsview isn't installed
            result = json.dumps(schema.jsonschema)

    output.write(result + "\n")
    output.flush()



from parsimonious import NodeVisitor
from collections import Sequence
from typing import Tuple, Optional, Set, Dict
import json

from . import tree as T
from .grammar import grammar
from .unspace import UnspaceVisitor


class TreeBuildingVisitor(NodeVisitor):

    SHELL_EXPRESSIONS = {
        "simple_type": 0,
        "card_content": 0,
        "card_1": (0, 0),
        "card_2": (0, -1),
        "object": 0,
        "object_property": 0,
        "object_pair_type": 0,
        "object_pair_name": 0,
        "array": 0,
        "parens": 1,
    }

    @staticmethod
    def unescape_string(escaped):
        return escaped.encode("utf-8").decode("unicode_escape")

    @staticmethod
    def gather_separated_list(first_item, other_items_with_separators) -> tuple:
        return (first_item,) + tuple(item[1] for item in other_items_with_separators)

    def visit_schema(self, node, c) -> T.Schema:
        value, definitions = c
        return T.Schema(value=value, definitions=definitions)

    def visit_sequence_and(self, node, c) -> T.Type:
        first, rest = c
        if len(rest):
            values = self.gather_separated_list(first, rest)
            return T.Operator(operator="allOf", values=values)
        else:
            return first

    def visit_type(self, node, c) -> T.Type:
        first, rest = c
        if len(rest) == 0:
            return first
        operands = self.gather_separated_list(first, rest)
        if all(isinstance(a, T.Constant) for a in operands):
            # Convert anyOf(const(...)...) into enum(...)
            constants = [a.value for a in operands]
            return T.Enum(values=constants)
        else:
            return T.Operator(operator="anyOf", values=operands)

    def visit_not_type(self, node, c) -> T.Not:
        return T.Not(value=c[1])

    def visit_string(self, node, c) -> T.String:
        return T.String(cardinal=c[1], format=None, regex=None)

    def visit_integer(self, node, c) -> T.Integer:
        _, cardinal, multiple = c
        return T.Integer(cardinal=cardinal, multiple=multiple)

    def visit_litteral(self, node, c) -> T.Litteral:
        # This rule is space-free
        return T.Litteral(value=node.children[0].text.lower())

    def visit_kw_forbidden(self, node, c) -> T.Forbidden:
        return T.Forbidden()

    def visit_lit_string(self, node, c) -> str:
        # This rule is space-free
        return node.text[1:-1]

    def visit_lit_regex(self, node, c) -> T.String:
        # This rule is space-free
        # Don't unescape the string
        return T.String(
            format=None, cardinal=(None, None), regex=node.children[-1].text[1:-1]
        )

    def visit_lit_format(self, node, c) -> T.String:
        # This rule is space-free
        # No need to unescape
        return T.String(
            cardinal=(None, None), regex=None, format=node.children[-1].text[1:-1]
        )

    def visit_opt_multiple(self, node, c) -> Optional[int]:
        return None if len(c) == 0 else c[0][1]

    def visit_opt_cardinal(self, node, c) -> Tuple[Optional[int], Optional[int]]:
        # generic_visit didn't detect card_1, probably due to some node optimization?
        if len(c) == 0:  # Empty cardinal
            return (None, None)
        uc = c[0][1]
        if isinstance(uc, int):
            return (uc, uc)
        else:  # Pair of ints/nulls
            return uc

    def visit_card_min(self, node, c) -> Tuple[int, None]:
        return (c[0], None)

    def visit_card_max(self, node, c) -> Tuple[None, int]:
        return (None, c[2])

    def visit_lit_integer(self, node, c) -> int:
        # This rule is space-free
        text = node.text
        base = 16 if len(text) > 2 and text[1] == "x" else 10
        return int(text, base)

    def visit_constant(self, node, c) -> T.Constant:
        # This rule is space-free
        source = node.text  # No need to unescape
        if source[0] == "`":
            assert source[-1] == "`"
            source = source[1:-1]
        try:
            value = json.loads(source)
            return T.Constant(value=value)
        except json.JSONDecodeError:
            raise ValueError(f"{source} is not a valid JSON fragment")

    def visit_object_keyword(self, node, c) -> T.Object:
        return T.Object(
            properties=[],
            additional_property_types=None,
            additional_property_names=None,
            cardinal=c[-1],
        )

    def visit_object_empty(self, node, c) -> T.Object:
        kwargs = {"properties": []}
        _, additional_props, _, kwargs["cardinal"] = c
        kwargs.update(additional_props)
        return T.Object(**kwargs)

    def visit_object_non_empty(self, node, c) -> T.Object:
        kwargs = {}
        _, additional_props, first_field, other_fields, _, kwargs["cardinal"] = c
        kwargs["properties"] = self.gather_separated_list(first_field, other_fields)
        kwargs.update(additional_props)
        return T.Object(**kwargs)

    def visit_object_pair(self, node, c) -> T.ObjectProperty:
        key, question, _, val = c
        if not isinstance(val, T.Type):  # wildcard
            val = None
        is_optional = bool(question) | isinstance(val, T.Forbidden)
        return T.ObjectProperty(self.unescape_string(key), is_optional, val)

    def visit_object_pair_unquoted_name(self, node, c) -> str:
        return node.text

    def visit_object_only(
        self, node, c
    ) -> Tuple[bool, Optional[T.Type], Optional[T.Type]]:
        """Parse `only`, `only <pattern>`, `only <pattern>: <type>` + optional coma."""
        if len(c) == 0:  # Empty sequence
            return {
                "additional_property_names": None,
                "additional_property_types": None,
            }
        _, maybe_something, _ = c[0]
        if len(maybe_something) == 0:  # keyword "only" alone
            return {
                "additional_property_names": None,
                "additional_property_types": False,
            }
        maybe_name, maybe_type = maybe_something[0]
        if maybe_name in ([], [[]]):  # No name, or wildcard name
            maybe_name = None
        else:
            maybe_name = maybe_name[0]
        if len(maybe_type):
            maybe_type = maybe_type[0][1]
        else:
            maybe_type = None
        return {
            "additional_property_names": maybe_name,
            "additional_property_types": maybe_type,
        }

    def visit_array_empty(self, node, c) -> T.Array:
        return T.Array(items=[], additional_items=True, cardinal=c[-1], unique=False)

    def visit_array_non_empty(self, node, c) -> T.Array:
        """
        With only one type and a "+" / "*" suffix, it's an homogeneous list.
        In this case, if there is a "only" qualifier it is ignored.
        With more than one type and a "+" / "*" suffix, it's a tuple type
        with extra items type. Here too, any "only" qualifier will be ignored.
        Without a suffix, it's a tuple type, the "only" qualifier is enforced.
        """
        _, array_prefix, first_item, other_items, extra, _, card = c

        items = self.gather_separated_list(first_item, other_items)

        if extra is None:  # No suffix -> tuple typing
            additional_items = "only" not in array_prefix
        elif extra == "+":
            additional_items = items[-1]
            # Don't remove it from required items, there must be at least one
        else:  # Last type is the type of extra items
            additional_items = items[-1]
            items = items[:-1]
        if card[0] is not None and len(items) >= card[0]:
            card = (None, card[1])  # Constraint is redundant
        if card[1] is not None and not additional_items and len(items) < card[1]:
            raise ValueError(
                f"An array cannot be both {len(items)} and <={card[1]} items long"
            )

        return T.Array(
            items=items,
            additional_items=additional_items,
            cardinal=card,
            unique="unique" in array_prefix,
        )

    def visit_array_prefix(self, node, c) -> Set[str]:
        """Return a set of strings among "unique" and "only"."""
        r = set()
        for word in ("unique", "only"):
            if word in node.text.lower():
                r.add(word)
        return r

    def visit_array_extra(self, node, c) -> str:
        # This rule is space-free
        if len(c) == 0:
            return None
        else:
            return node.children[0].text

    def visit_conditional(self, node, c):
        _, if_term, _, then_term, elif_parts, else_part = c
        else_term = else_part[0][1] if len(else_part) else None
        elif_terms = [[cond, val] for (_, cond, _, val) in elif_parts]

        def rec(c, t, elifs, e):
            """Change elif pairs into nested if/then/else."""
            if elifs:
                (c2, t2) = elifs[-1]
                return rec(c, t, elifs[:-1], rec(c2, t2, [], e))
            else:
                return T.Conditional(if_term=c, then_term=t, else_term=e)

        return rec(if_term, then_term, elif_terms, else_term)

    def visit_opt_definitions(self, node, c) -> T.Definitions:
        if len(c) == 0:  # Empty definition
            return T.Definitions(values={})
        else:
            _, defs = c[0]
            return defs

    def visit_definitions(self, node, c) -> T.Definitions:
        first_def, other_defs_with_and = c
        other_defs = [d[1] for d in other_defs_with_and]
        items = (first_def, *other_defs)
        return T.Definitions(values=dict(items))

    def visit_definition(self, node, c) -> Tuple[str, T.Type]:
        _, id, _, _, type = c
        return (id, type)

    def visit_def_identifier(self, node, c) -> str:
        return node.text

    def visit_def_reference(self, node, c) -> T.Reference:
        return T.Reference(value=c[1])

    def generic_visit(self, node, c) -> tuple:
        """ The generic visit method. """
        n = self.SHELL_EXPRESSIONS.get(node.expr_name)
        if n is None:
            return c
        elif isinstance(n, Sequence):
            return tuple(c[i] for i in n)
        elif isinstance(n, int):
            return c[n]
        else:
            raise ValueError(f"bad SHELL_EXPRESSIONS for {node.expr_name}")


jscn_visitor = TreeBuildingVisitor()
unspace_visitor = UnspaceVisitor()


def parse(what: str, source: str, verbose=False) -> T.Type:
    raw_tree = grammar[what].parse(source)
    unspaced_tree = unspace_visitor.visit(raw_tree)
    if verbose:
        print("PEG tree:\n" + unspaced_tree.prettily())
    parsed_tree = jscn_visitor.visit(unspaced_tree)
    if verbose:
        print("JSCN tree:\n" + parsed_tree.prettily())
    parsed_tree.source = source
    return parsed_tree



"""PEG Grammar for JSON-Schema compact notation."""
import re
from parsimonious import Grammar

src = r"""
schema = _ type _ opt_definitions _
type = sequence_and (_ or _ sequence_and)*
sequence_and = simple_type (_ and _ simple_type)*
simple_type = litteral / kw_forbidden / string / object / integer / array /
       lit_regex / lit_format / constant / parens / not_type / def_reference /
       conditional
parens = lparen _ type _ rparen

litteral = KEYWORD("boolean") / KEYWORD("null") / KEYWORD("number")

lit_integer =  ~"0x[0-9a-fA-F]+" / ~"[0-9]+"
lit_regex = regex_prefix lit_string
lit_format = format_prefix lit_string

lit_string = quote_char (non_quote_char*) quote_char

quote_char = "\""
backslash_char = "\\"
non_quote_char = escaped_char / ~"[^\"]"
escaped_char = backslash_char ~"."

constant = backquote_constant / lit_string
backquote_constant = backquote_char (neither_quote_nor_backquote / lit_string)* backquote_char
backquote_char = "`"
neither_quote_nor_backquote = escaped_char / (~"[^\"`]")

string = KEYWORD("string") _ opt_cardinal
integer = KEYWORD("integer") _ opt_cardinal _ opt_multiple
opt_multiple = ("/" _ lit_integer)?

not_type = kw_not _ simple_type

regex_prefix = KEYWORD("r")
format_prefix = KEYWORD("f")
wildcard = "_"
or = "|"
and = "&"
lparen = "("
rparen = ")"
lbrace = "{"
rbrace = "}"
lbracket = "["
rbracket = "]"
comma = ","
colon = ":"
question = "?"
star = "*"
plus = "+"

kw_array = KEYWORD("array")
kw_forbidden = KEYWORD("forbidden")
kw_if = KEYWORD("if")
kw_then = KEYWORD("then")
kw_elif = KEYWORD("elif")
kw_else = KEYWORD("else")
kw_not = KEYWORD("not")
kw_object = KEYWORD("object")
kw_only = KEYWORD("only")
kw_unique = KEYWORD("unique")
kw_where = KEYWORD("where")
kw_and = KEYWORD("and")

conditional =
    kw_if _ type _ kw_then _ type
    (_ kw_elif _ type _ kw_then _ type)*
    (_ kw_else _ type)?

opt_cardinal = (lbrace _ card_content _ rbrace)?
card_content = card_2 / card_min / card_max / card_1
card_2 = lit_integer _ comma _ lit_integer
card_1 = lit_integer
card_min = lit_integer _ comma? _ wildcard
card_max = wildcard _ comma? _ lit_integer

object = object_empty / object_non_empty / object_keyword
object_empty = lbrace _ object_only _ rbrace opt_cardinal
object_keyword = kw_object opt_cardinal
object_non_empty = lbrace _
                   object_only _
                   object_pair (_ comma _ object_pair)* _
                   rbrace
                   opt_cardinal
object_only = (kw_only _ ((lit_regex/def_reference/wildcard) _ (colon _ type)?)? comma?)?
object_pair = object_pair_name _ question? _ colon _ object_pair_type
object_pair_name = lit_string / object_pair_unquoted_name
object_pair_unquoted_name = ~"[A-Za-z0-9][-_A-Za-z0-9]*"
object_pair_type = type / wildcard

array = array_empty / array_non_empty
array_empty = ((lbracket _ rbracket) / kw_array) _ opt_cardinal
array_non_empty = lbracket _ array_prefix _
                  type (_ comma _ type)* _
                  array_extra _
                  rbracket _
                  opt_cardinal
array_prefix = ((kw_only / kw_unique) _) *
array_extra = (plus / star)?

opt_definitions = (kw_where _ definitions)?
definitions = _ definition (_ kw_and _ definition)* _
definition = "<"? def_identifier ">"? _ def_equal _ type
def_equal = "="
def_reference = "<" def_identifier ">"
def_identifier = ~"[A-Za-z_][-A-Za-z_0-9]*"

_ = meaninglessness*
meaninglessness = ~r"\s+" / comment
comment = ~r"#[^\r\n]*"
"""


def parse_keywords(src):
    """Replace occurences of KEYWORD("...") by a regex allowing all capitalizations
    of the keyword between quotes."""

    def f(m):
        return '~"' + "".join(f"[{x}{x.upper()}]" for x in m[1]) + '"'

    r = re.compile(r'KEYWORD\("([a-z]+)"\)')
    return r.sub(f, src)


grammar = Grammar(parse_keywords(src))



from parsimonious import NodeVisitor


class UnspaceVisitor(NodeVisitor):
    """Remove space tokens from a Parsimonious parse tree."""

    # TODO generate a set of space offsets in the code, so that they can
    # be compacted properly, including comments and excluding string contents.

    WHITESPACE_TOKEN = object()  # Sentinel value

    def visit__(self, node, children) -> object:
        """Replace whitspaces with an easy-to-filter sentinel value."""
        return self.WHITESPACE_TOKEN

    def generic_visit(self, node, children) -> tuple:
        """ The generic visit method. """
        node.children = [x for x in children if x is not self.WHITESPACE_TOKEN]
        return node



import unittest
from . import Schema, Definitions
import json
import jsonschema
from . import tree as T
from parsimonious.exceptions import IncompleteParseError
import re


class TestJSCN(unittest.TestCase):
    def cmp(self, src, sch: dict) -> None:
        if isinstance(src, str):
            sch2 = Schema(src).jsonschema
        elif isinstance(src, T.Schema):
            sch2 = src.jsonschema
        deleted_keys = ["$comment", "$schema"]
        for k in deleted_keys:
            del sch2[k]
        self.assertDictEqual(sch2, sch)

    def test_simple(self):
        self.cmp("boolean", {"type": "boolean"})
        self.cmp("string", {"type": "string"})
        self.cmp("integer", {"type": "integer"})
        self.cmp("number", {"type": "number"})
        self.cmp("null", {"type": "null"})

    def test_lit_string(self):
        self.cmp('r"foo"', {"type": "string", "pattern": "foo"})
        self.cmp('f"foo"', {"type": "string", "format": "foo"})
        self.cmp('r"foo\\"bar"', {"type": "string", "pattern": r"foo\"bar"})
        self.cmp('r"foo\\""', {"type": "string", "pattern": r"foo\""})

    def test_integer(self):
        self.cmp("integer{_,1}", {"type": "integer", "maximum": 1})
        self.cmp("integer{1,_}", {"type": "integer", "minimum": 1})
        self.cmp("integer{0,10}", {"type": "integer", "minimum": 0, "maximum": 10})
        self.cmp("integer/5", {"type": "integer", "multipleOf": 5})

    def test_string(self):
        self.cmp("string{_,1}", {"type": "string", "maxLength": 1})
        self.cmp("string{1,_}", {"type": "string", "minLength": 1})
        self.cmp("string{0,10}", {"type": "string", "minLength": 0, "maxLength": 10})

    def test_constant(self):
        self.cmp("`123`", {"const": 123})
        self.cmp('"123"', {"const": "123"})
        self.cmp('`"123"`', {"const": "123"})
        self.cmp('`{"a": 1}`', {"const": {"a": 1}})

    def test_constant_with_backquote(self):
        self.cmp(r'`"`"`', {"const": "`"})
        with self.assertRaises(IncompleteParseError):
            Schema('`{"foo": `}`')

    def test_enum(self):
        self.cmp("`1`|`2`|`3`", {"enum": [1, 2, 3]})
        self.cmp('"1"|"2"|"3"', {"enum": ["1", "2", "3"]})

    def test_forbidden(self):
        self.assertEqual(Schema("forbidden").jsonschema, False)
        self.assertDictEqual(Schema("{x: forbidden}").jsonschema, Schema("{x?: forbidden}").jsonschema)
        self.assertNotEqual(Schema("{x: `1`}").jsonschema, Schema("{x?: `1`}").jsonschema)

    def test_object_empty(self):
        obj = {"type": "object"}
        self.cmp("{}", obj)
        self.cmp("object", obj)

    def test_object_simple(self):
        self.cmp(
            "{only foo: integer}",
            {
                "type": "object",
                "required": ["foo"],
                "properties": {"foo": {"type": "integer"}},
                "additionalProperties": False,
            },
        )
        self.cmp(
            "{foo: integer}",
            {
                "type": "object",
                "required": ["foo"],
                "properties": {"foo": {"type": "integer"}},
            },
        )
        self.cmp(
            "{foo?: integer}",
            {"type": "object", "properties": {"foo": {"type": "integer"}}},
        )
        self.cmp(
            '{"foo"?: integer}',
            {"type": "object", "properties": {"foo": {"type": "integer"}}},
        )
        self.cmp(
            "{_: integer}",
            {"type": "object", "additionalProperties": {"type": "integer"}},
        )

    def test_object_card(self):
        pass  # TODO

    def test_array_empty(self):
        array = {"type": "array"}
        self.cmp("[]", array)
        self.cmp("array", array)

    def test_object_simple(self):
        integer = {"type": "integer"}
        self.cmp(
            "[only integer]",
            {"type": "array", "items": [integer], "additionalItems": False},
        )
        self.cmp("[integer]", {"type": "array", "items": [integer]})
        self.cmp("[integer*]", {"type": "array", "items": integer})  # List-notation rather than tuple notation
        self.cmp(
            "[integer+]",  # Maybe encoded either that way or with a cardinal
            {"type": "array", "items": [integer], "additionalItems": integer},
        )

    def test_array_card(self):
        pass  # TODO

    def test_where(self):
        s = '{only <id>: <byte>} where id = r"[a-z]+" and byte = integer{0,0xFF}'
        self.cmp(
            s,
            {
                "additionalProperties": {"$ref": "#/definitions/byte"},
                "definitions": {
                    "byte": {"maximum": 255, "minimum": 0, "type": "integer"},
                    "id": {"pattern": "[a-z]+", "type": "string"},
                },
                "propertyNames": {"$ref": "#/definitions/id"},
                "type": "object",
            },
        )

    def test_x0(self):
        Schema(
            r"""{
            kind: `"aircraft"`,
            mission: string
        } | {
            kind: `"mission"`,
            name: string,
            fleet: {only <id>: string}
        } where id = r"[a-z]+" """
        ).to_jsonschema()

    def test_x1(self):
        Schema(
            """
        { instance: <plid>,
          ground: <plid>,
          mission: <plid>,
          fleet: { only <plid>: <aircraft> }
        }
        where plid = r"[A-Z0-9]{4}"
        and aircraft = {
          color?: string,
          status?: `"online"` | `"offline"`
        }
        """
        ).to_jsonschema()

    def test_prune(self):
        s = Schema(
            """
        {prop: <used_1>}
        where used_1 = [<used_2>+]
        and unused_1 = [<unused_2>+]
        and used_2 = integer
        and unused_2 = string
        """
        )
        self.assertSetEqual(set(s.jsonschema["definitions"].keys()), {"used_1", "used_2"})

    def test_combine_1(self):
        s = Schema("""{prop: <used_1>}""")
        d = Definitions(
            """
        used_1 = [<used_2>+]
        and unused_1 = [<unused_2>+]
        and used_2 = integer
        and unused_2 = string
        """
        )
        s |= d
        self.assertSetEqual(set(s.jsonschema["definitions"].keys()), {"used_1", "used_2"})

    def test_combine_2(self):
        s = (
            Schema("{prop: <used_1>}")
            | Definitions("used_1 = [<used_2>+]")
            | Definitions("unused_1 = [<unused_2>+]")
            | Definitions("used_2 = integer")
            | Definitions("unused_2 = string")
        )
        self.assertSetEqual(set(s.jsonschema["definitions"].keys()), {"used_1", "used_2"})

    def test_combine_3(self):
        """combine defs together before adding the result to the schema."""
        s = Schema("{prop: <used_1>}") | (
            Definitions("used_1 = [<used_2>+]")
            | Definitions("unused_1 = [<unused_2>+]")
            | Definitions("used_2 = integer")
            | Definitions("unused_2 = string")
        )
        self.assertSetEqual(set(s.jsonschema["definitions"].keys()), {"used_1", "used_2"})

    def test_combine_schemas(self):
        self.cmp(
            "{foo?: integer} & {bar?: string}",
            {
                "allOf": [
                    {"properties": {"foo": {"type": "integer"}}, "type": "object"},
                    {"properties": {"bar": {"type": "string"}}, "type": "object"},
                ]
            },
        )

    def test_missing_def(self):
        with self.assertRaisesRegex(ValueError, "Missing definition"):
            Schema("{foo: <bar>}").to_jsonschema()

    def test_validate(self):
        s = Schema("{only x: integer}")
        s.validate()
        s.validate({"x": 1})
        with self.assertRaises(jsonschema.ValidationError):
            s.validate({"x": "1"})
        with self.assertRaises(jsonschema.ValidationError):
            s.validate({"x": 1, "y": 2})

    def test_def_conflict(self):
        with self.assertRaisesRegex(ValueError, "conflict"):
            Definitions("x = integer") | Definitions("x = string")

    def test_toplevel_ref(self):
        self.cmp(
            "<x> where x=integer",
            {"$ref": "#/definitions/x", "definitions": {"x": {"type": "integer"}}},
        )
        s = Schema("<x>")
        d = Definitions("x=integer")
        self.cmp(
            s | d,
            {"$ref": "#/definitions/x", "definitions": {"x": {"type": "integer"}}},
        )

    def test_indirect_ref(self):
        self.cmp(
            '<A> where A = <B> and B = `"C"`',
            {
                "$ref": "#/definitions/A",
                "definitions": {"A": {"$ref": "#/definitions/B"}, "B": {"const": "C"}},
            },
        )

    def test_cond_1(self):
        self.cmp(
            r'''
            if {country: `"USA"`} then {code: <us_postcode>}
            elif {country: `"Canada"`} then {code: <ca_postcode>}
            elif {country: `"NL"`} then {code: <nl_postcode>}
            else {code: string}
            where us_postcode = r"\d{5}(-\d{4})?"
            and ca_postcode = r"[A-Z]\d[A-Z] \d[A-Z]\d"
            and nl_postcode = r"[0-9]{4} [A-Z]{2}"''',
            {
                "if": {
                    "type": "object",
                    "required": ["country"],
                    "properties": {"country": {"const": "USA"}},
                },
                "then": {
                    "type": "object",
                    "required": ["code"],
                    "properties": {"code": {"$ref": "#/definitions/us_postcode"}},
                },
                "else": {
                    "if": {
                        "type": "object",
                        "required": ["country"],
                        "properties": {"country": {"const": "Canada"}},
                    },
                    "then": {
                        "type": "object",
                        "required": ["code"],
                        "properties": {"code": {"$ref": "#/definitions/ca_postcode"}},
                    },
                    "else": {
                        "if": {
                            "type": "object",
                            "required": ["country"],
                            "properties": {"country": {"const": "NL"}},
                        },
                        "then": {
                            "type": "object",
                            "required": ["code"],
                            "properties": {"code": {"$ref": "#/definitions/nl_postcode"}},
                        },
                        "else": {
                            "type": "object",
                            "required": ["code"],
                            "properties": {"code": {"type": "string"}},
                        },
                    },
                },
                "definitions": {
                    "us_postcode": {"type": "string", "pattern": r"\d{5}(-\d{4})?"},
                    "ca_postcode": {
                        "type": "string",
                        "pattern": r"[A-Z]\d[A-Z] \d[A-Z]\d",
                    },
                    "nl_postcode": {"type": "string", "pattern": r"[0-9]{4} [A-Z]{2}"},
                },
            },
        )

    def test_cond_2(self):
        self.cmp(
            r'''
            if {country: `"USA"`} then {code: <us_postcode>}
            else {code: string}
            where us_postcode = r"\d{5}(-\d{4})?"''',
            {
                "if": {
                    "type": "object",
                    "required": ["country"],
                    "properties": {"country": {"const": "USA"}},
                },
                "then": {
                    "type": "object",
                    "required": ["code"],
                    "properties": {"code": {"$ref": "#/definitions/us_postcode"}},
                },
                "else": {
                    "type": "object",
                    "required": ["code"],
                    "properties": {"code": {"type": "string"}},
                },
                "definitions": {"us_postcode": {"type": "string", "pattern": r"\d{5}(-\d{4})?"}},
            },
        )

    def test_cond_3(self):
        self.cmp(
            r'''
            if {country: `"USA"`} then {code: <us_postcode>}
            where us_postcode = r"\d{5}(-\d{4})?"''',
            {
                "if": {
                    "type": "object",
                    "required": ["country"],
                    "properties": {"country": {"const": "USA"}},
                },
                "then": {
                    "type": "object",
                    "required": ["code"],
                    "properties": {"code": {"$ref": "#/definitions/us_postcode"}},
                },
                "definitions": {"us_postcode": {"type": "string", "pattern": r"\d{5}(-\d{4})?"}},
            },
        )

    def test_capitalized_keywords(self):
        Schema('{ONLY <foo>: <bar>} WHERE foo = r"^[0-9]+$" AnD bar= r"^[a-z]+$"')
        Schema("[UNIQUE integer+]")
        with self.assertRaisesRegex(ValueError, "Missing definition"):
            Schema("{x: <foo>} WHERE FOO=integer").to_jsonschema()

    def str_check(self, src: str):
        s = Schema(src)
        src2 = str(s)
        src = re.sub(r"\s+", "", src)
        src2 = re.sub(r"\s+", "", src2)
        self.assertEqual(src, src2)

    def test_str_cardinal(self):
        self.str_check("integer")
        self.str_check("integer{_, 123}")
        self.str_check("integer{123, _}")
        self.str_check("integer{1, 23}")

    def test_str_object(self):
        self.str_check("{}")
        self.str_check("{only}")
        self.str_check("{only foo: string}")
        self.str_check('{only r"[a-z]+"}')
        self.str_check('{only r"[a-z]+": integer}')
        self.str_check('{only r"[a-z]+", foo: string}')
        self.str_check("{foo: string}")
        self.str_check('{"foo!": string}')
        self.str_check(r'{"a\"b": string}')

    def test_str_array(self):
        self.str_check("[]")
        self.str_check("[only boolean]")
        self.str_check("[boolean*]")
        self.str_check("[only unique boolean]")
        self.str_check("[only boolean, integer]")
        self.str_check("[only boolean, integer, integer]")

    def test_str_schema(self):
        self.str_check("<x> where x=`1`")
        self.str_check("<x>|<y> where x=`1` and y=`2`")

    def test_str_str(self):
        self.str_check("string")
        self.str_check('r"123"')
        self.str_check('f"123"')

    def test_only_wildcard(self):
        self.cmp("{only _: {}}", {"type": "object", "additionalProperties": {"type": "object"}})

    def test_wildcard_props(self):
        self.cmp(
            "{only x: string, y?: _}",
            {
                "type": "object",
                "required": ["x"],
                "properties": {"x": {"type": "string"}, "y": True},
                "additionalProperties": False,
            },
        )


if __name__ == "__main__":
    unittest.main()



import re
from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()


with open("jsonschema_cn/__init__.py") as fh:
    version = re.search(r'^__version__\s*=\s*["\'](.*)["\']', fh.read(), re.M).group(1)


setup(name='jsonschema_cn',
      version=version,
      description='Compact notation for JSON Schemas',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://github.com/fab13n/jsonschema_cn',
      author='Fabien Fleutot',
      author_email='fleutot@gmail.com',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent",
      ],
      keywords='DSL JSON schema jsonschema',
      license='BSD',
      packages=['jsonschema_cn'],
      install_requires=[
          'parsimonious>=0.8.0',
          'jsonschema>=3.0.1'
      ],
      entry_points={
          "console_scripts": ['jscn = jsonschema_cn.cli:main']
      },
)
